/* Copyright (c) 2014 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
*
* @defgroup gpiote_example_main main.c
* @{
* @ingroup nrf_gpiote_example
* @brief GPIOTE Example Application main file.
*
* This file contains the source code for a sample application using GPIOTE.
*/

#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>

#include "nrf.h"
#include "nrf_gpiote.h"
#include "nrf_gpio.h"
#include "boards.h"
#include "nrf_drv_ppi.h"
//#include "nrf_drv_timer.h"
#include "nrf_drv_gpiote.h"
#include "nrf_drv_spi.h"
#include "app_error.h"

#define NRF_LOG_MODULE_NAME "TEST"
#include "nrf_log.h"
#include "nrf_log_ctrl.h"

#ifdef BSP_LED_0
    #define GPIO_OUTPUT_PIN_NUMBER BSP_LED_0  /**< Pin number for output. */
#endif
#ifndef GPIO_OUTPUT_PIN_NUMBER
    #error "Please indicate output pin"
#endif

#define MPU_SPI_BUFFER_SIZE 14
#define MPU_SPI_TIMEOUT 2000
#define MPU_SPI_WRITE_BIT       0x00
#define MPU_SPI_READ_BIT        0x80

static const nrf_drv_spi_t m_spi_instance = NRF_DRV_SPI_INSTANCE(0);
volatile static bool spi_tx_done = false;


uint8_t spi_tx_buffer[MPU_SPI_BUFFER_SIZE];
uint8_t spi_rx_buffer[MPU_SPI_BUFFER_SIZE];

// PPI resources
nrf_ppi_channel_t ppi_channel_toggle;
nrf_ppi_channel_t ppi_channel_int;

void mpu_spi_event_handler(const nrf_drv_spi_evt_t *evt)
{
    if(evt->type == NRF_DRV_SPI_EVENT_DONE)
    {
        NRF_LOG_DEBUG("TxDone\n");
        NRF_LOG_HEXDUMP_INFO(spi_rx_buffer,6);
        spi_tx_done = true;
    }
    else
    {
        NRF_LOG_DEBUG("Wrong Event\n");
        // Something is wrong
    }
}

uint32_t mpu_spi_init(void)
{
    
    const nrf_drv_spi_config_t spi_mpu_config = {                                                            \
        .sck_pin      = SPI0_CONFIG_SCK_PIN, 
        .mosi_pin     = SPI0_CONFIG_MOSI_PIN,
        .miso_pin     = SPI0_CONFIG_MISO_PIN,
        .ss_pin       = SPI0_CONFIG_SS_PIN,
        .irq_priority = SPI_DEFAULT_CONFIG_IRQ_PRIORITY, 
        .orc          = 0xFF,                                    
        .frequency    = NRF_DRV_SPI_FREQ_1M,                     
        .mode         = NRF_DRV_SPI_MODE_3,                      
        .bit_order    = NRF_DRV_SPI_BIT_ORDER_MSB_FIRST,         
    };
    return nrf_drv_spi_init(&m_spi_instance, &spi_mpu_config, mpu_spi_event_handler);  
}


static void ppi_setup()
{
    uint32_t int_evt_addr;
    uint32_t spi_evt_addr;
    uint32_t spi_task_addr;
    uint32_t gpiote_task_addr;
    ret_code_t err_code;
    nrf_drv_gpiote_out_config_t config_out = GPIOTE_CONFIG_OUT_TASK_TOGGLE(false);

    err_code = nrf_drv_gpiote_out_init(GPIO_OUTPUT_PIN_NUMBER, &config_out);
    APP_ERROR_CHECK(err_code);

    // Setup the input interrupt pin
    nrf_drv_gpiote_in_config_t config_in = GPIOTE_CONFIG_IN_SENSE_LOTOHI(true);
    err_code = nrf_drv_gpiote_in_init(MPU_INT_PIN, &config_in, NULL); // No handler function
    APP_ERROR_CHECK(err_code);
    
    // allocate PPI channels for hardware
    err_code = nrf_drv_ppi_channel_alloc(&ppi_channel_int);
    APP_ERROR_CHECK(err_code);

    err_code = nrf_drv_ppi_channel_alloc(&ppi_channel_toggle);
    APP_ERROR_CHECK(err_code);

    int_evt_addr = nrf_drv_gpiote_in_event_addr_get(MPU_INT_PIN);
    spi_task_addr = nrf_drv_spi_start_task_get(&m_spi_instance);
    err_code = nrf_drv_ppi_channel_assign(ppi_channel_int, int_evt_addr, spi_task_addr);
    APP_ERROR_CHECK(err_code);

    spi_evt_addr = nrf_drv_spi_end_event_get(&m_spi_instance);
    gpiote_task_addr = nrf_drv_gpiote_out_task_addr_get(GPIO_OUTPUT_PIN_NUMBER);

    err_code = nrf_drv_ppi_channel_assign(ppi_channel_toggle, spi_evt_addr, gpiote_task_addr);
    APP_ERROR_CHECK(err_code);
}

void transfer_enable()
{
    ret_code_t err_code;
    // end to front
    nrf_drv_gpiote_out_task_enable(GPIO_OUTPUT_PIN_NUMBER);
    err_code = nrf_drv_ppi_channel_enable(ppi_channel_toggle);
    APP_ERROR_CHECK(err_code);
    err_code = nrf_drv_ppi_channel_enable(ppi_channel_int);
    APP_ERROR_CHECK(err_code);
    nrf_drv_gpiote_in_event_enable(MPU_INT_PIN,true);
}

void transfer_disable()
{
    ret_code_t err_code;
    
    nrf_drv_gpiote_in_event_disable(MPU_INT_PIN);
    err_code = nrf_drv_ppi_channel_disable(ppi_channel_int);
    APP_ERROR_CHECK(err_code);
    err_code = nrf_drv_ppi_channel_disable(ppi_channel_toggle);
    APP_ERROR_CHECK(err_code);
    nrf_drv_gpiote_out_task_disable(GPIO_OUTPUT_PIN_NUMBER);
}

uint32_t mpu_read_registers(uint8_t reg, uint8_t * p_data, uint32_t length)
{
    uint32_t err_code;
    uint32_t timeout = MPU_SPI_TIMEOUT;
//    uint32_t timeout = 0;
    
    // Add read bit to register. 
    reg = reg | MPU_SPI_READ_BIT;
	
    // Read data over SPI and store incomming data in spi_rx_buffer
    err_code = nrf_drv_spi_transfer(&m_spi_instance, &reg, 1, spi_rx_buffer, length + 1); // Length + 1 because register byte has to be clocked out before MPU returns data of length 'length'
    if(err_code != NRF_SUCCESS) return err_code;
    NRF_LOG_DEBUG("Waiting for SPI Xfer\n")

    while((!spi_tx_done) && --timeout);
    NRF_LOG_DEBUG("Timeout: %d\n", timeout);
    if(!timeout) return NRF_ERROR_TIMEOUT;
    spi_tx_done = false;
	
  	// Copy data in spi_rx_buffer over to p_data
	 memcpy(p_data, &spi_rx_buffer[1], length);

    return NRF_SUCCESS;
}

uint32_t mpu_write_registers(uint8_t reg, uint8_t * p_data, uint32_t length)
{
    uint32_t err_code;
    
    if(length > MPU_SPI_BUFFER_SIZE - 1) // Must be space for register byte in buffer
    {
        return NRF_ERROR_DATA_SIZE;
    }
    
    uint32_t timeout = MPU_SPI_TIMEOUT;

    // Add write bit to register. 
    reg = reg | MPU_SPI_WRITE_BIT;
    
    spi_tx_buffer[0] = reg;
    memcpy((spi_tx_buffer + 1), p_data, length);
    
    err_code = nrf_drv_spi_transfer(&m_spi_instance, spi_tx_buffer, length + 1, NULL, 0);
    if(err_code != NRF_SUCCESS) return err_code;
    NRF_LOG_DEBUG("Waiting for SPI Xfer\n")

    while((!spi_tx_done) && --timeout);
    NRF_LOG_DEBUG("Timeout: %d\n", timeout);
    if(!timeout) return NRF_ERROR_TIMEOUT;
    spi_tx_done = false;

    return err_code;
}

void mpu_init()
{
    uint32_t err_code;
    uint8_t loc_buffer[MPU_SPI_BUFFER_SIZE];
    
    loc_buffer[0] = 0x11; // Disable I2C Slave interface (SPI only) Resets gyro, accelerometer and temperature sensor signal paths.
    loc_buffer[1] = 1; // Resets gyro, accelerometer and temperature sensor signal paths.
    NRF_LOG_INFO("Resetting signal path\n");
    err_code = mpu_write_registers(0x6A, loc_buffer, 2);
    APP_ERROR_CHECK(err_code);
    // Chose  PLL with X axis gyroscope reference as clock source
//    NRF_LOG_DEBUG("Setting clock source\n");
//    err_code = mpu_write_registers(0x6B, loc_buffer, 1);
//    APP_ERROR_CHECK(err_code);
    NRF_LOG_FLUSH();
    err_code = mpu_read_registers(0x75, loc_buffer, 1);
    APP_ERROR_CHECK(err_code);
    NRF_LOG_INFO("WHO AM I: %0x\n", loc_buffer[0]);

    NRF_LOG_FLUSH();
    // Setup MPU for operation
    loc_buffer[0] = 249; // Sample rate div -> 1000/(1+SRD)
    loc_buffer[1] = 0x01; // config: b0:fifo(1)=0:ext_sync(3)=0:DLPF(3)=1
    loc_buffer[2] = 0x38; // Gyro Config: xgt_en(1)=0:ygt_en(1)=0:zgt_en(1)=0:GF_sel(2)=3:-(1)=0:fchoice(2)=0
    loc_buffer[3] = 0x00; // Accel Config: xst_en(1)=0:yst_en(1)=0:zst_en(1)=0:AF_sel(2)=0:-(3)=0
    loc_buffer[4] = 0x03; // Accel Config2: -(4)=0:Afchoice(1)=0:adlpf(3)=3
    NRF_LOG_INFO("Setting Accel and Gyro config\n");
    err_code = mpu_write_registers(0x19, loc_buffer, 5);
    APP_ERROR_CHECK(err_code);

    NRF_LOG_FLUSH();
    // Configure Interrupts     
    loc_buffer[0] = 0x10; // Interrupt config: actl(1)=0:open(1)=0:latch(1)=1:clr_any(1)=1:actl_fsync(1)=0:fsync_int(1)=0:byp_en(1)=0:-(1)=0 
    NRF_LOG_INFO("Configuring interrupts\n");
    err_code = mpu_write_registers(0x37, loc_buffer, 1);
    APP_ERROR_CHECK(err_code);
    NRF_LOG_FLUSH();
}

void mpu_enable_interrupts()
{
    uint8_t loc_buffer[MPU_SPI_BUFFER_SIZE];
    loc_buffer[0] = 0x01; // Interrupt enable: -(1)=0:wom(1)=0:-(1)=0:f_ovr(1)=0:fsync(1)=0:-(2)=0:d_rdy(1)=1 
    NRF_LOG_INFO("Enabling interrupts\n");
    mpu_write_registers(0x38, loc_buffer, 1);
}

void mpu_disable_interrupts()
{
    uint8_t loc_buffer[MPU_SPI_BUFFER_SIZE];
    loc_buffer[0] = 0x00; // Interrupt enable: -(1)=0:wom(1)=0:-(1)=0:f_ovr(1)=0:fsync(1)=0:-(2)=0:d_rdy(1)=0 
    NRF_LOG_INFO("Disabling interrupts\n");
    mpu_write_registers(0x38, loc_buffer, 1);
    mpu_read_registers(0x39, loc_buffer,1); // Read the interrupt status register to clear any pending interrupts
}

void mpu_prepare_transfers()
{
    spi_tx_buffer[0] = 0x3B;
    memset(spi_rx_buffer,0,MPU_SPI_BUFFER_SIZE);
    nrf_drv_spi_xfer_desc_t xfer = NRF_DRV_SPI_XFER_TRX(spi_tx_buffer, 1, spi_rx_buffer, 6);
    uint32_t flags = NRF_DRV_SPI_FLAG_HOLD_XFER           |
                     NRF_DRV_SPI_FLAG_REPEATED_XFER;
    nrf_drv_spi_xfer(&m_spi_instance, &xfer, flags);    
}        

/**
 * @brief Function for application main entry.
 */
int main(void)
{
    int16_t x;
    int16_t y;
    int16_t z;
    ret_code_t err_code;
    err_code = NRF_LOG_INIT(NULL);
    APP_ERROR_CHECK(err_code);

    err_code = nrf_drv_ppi_init();
    APP_ERROR_CHECK(err_code);

    err_code = nrf_drv_gpiote_init();
    APP_ERROR_CHECK(err_code);

    err_code = mpu_spi_init();
    APP_ERROR_CHECK(err_code);
    // Setup PPI channel with event from TIMER compare and task GPIOTE pin toggle.
    ppi_setup();

    NRF_LOG_FLUSH();
    // Setup the MPU for interrupt driven events
    mpu_init();
    NRF_LOG_FLUSH();
    
    mpu_enable_interrupts(); // Before we setup continuous SPI transfers
    mpu_prepare_transfers();
    // Enable transfers
    transfer_enable();

    while (true)
    {
        // Do Nothing - GPIO can be toggled without software intervention.
        if (spi_tx_done) {
          x = (spi_rx_buffer[1] << 8) | spi_rx_buffer[2];
          y = (spi_rx_buffer[3] << 8) | spi_rx_buffer[4];
          z = (spi_rx_buffer[5] << 8) | spi_rx_buffer[6];
//          nrf_drv_gpiote_out_task_trigger(GPIO_OUTPUT_PIN_NUMBER);
          NRF_LOG_INFO("X: %d,Y: %d, Z: %d\n",x,y,z);
          spi_tx_done = false;
        }
        NRF_LOG_PROCESS();
    }
}


/** @} */
